/***
*		space.c -- a real manly panning tool for b&p
*		 1993, 1994 Another Option
***/

/*
============================================================================
|    DATE   | WHAT
============================================================================
	 6 Jan 93 | created from panner.c
	 7 Jan 93 | fixed (and kludged) gain()
	 9 Jan 93 | removed support for speaker-spread angle: do it with scaling 
	          | the x inputs
   10 Jan 93 | added sysex handling
   25 Jan 93 | massaged for sas/c v6.0
   14 Jan 93 | massaged for bplib.lib
	13 Nov 93 | added snapshot & new window title
	14 Nov 93 | added sysex support for controller numbers (even though it's likely useless)
	25 Jan 94 | new ui; new options for passing inputs and events; modified sysex
	          | format: separate bytes for trash and dupe instead of packing
   12 Nov 96 | doing the regargs thing
*/

#define DBUG 0

#define THIS_TOOL_NAME "SPAS"

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <string.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <stddef.h>
//#include <stdio.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
#include <stdlib.h>
#include <math.h>

#include "bars.h"
#include "barsmacros.h"
#include "bplib.h"
#include "snapshotgadget.h"
#include "proto/space_protos.h"
#include "space2w.h"
#include "/icons/spaceup.h"
#include "/icons/spacedown.h"
#include "debug.h"

extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
static UBYTE const versiontag[] = "$VER: Space 37.15 (12.11.96)";

/*
*  note that angle and theta are still in the tool structure for upward
*  compatability (hah)
*/
struct SpaceTool {		/* a Tool structure for the tool */
	struct Tool tool;
	UBYTE in;			/* controller to interpret as virtual x location */
	UBYTE distance;	/* controller to read as virtual y location */
	UBYTE angle;		/* controller to use as speaker separation angle */
	UBYTE tout;			/* controller to send on master track */
	UBYTE bout;			/* controller to send on branch track */
	UBYTE master;		/* which track gets to be "left"? */
	UBYTE trash; 		/* free the incoming control message? */
	UBYTE dupe;			/* duplicate notes and send to branch? */
	double x;			/* last virtual x coordinate */
	double y;			/* last virtual y coordinate */
	double theta;		/* last speaker-spread angle value (radians) */
};

/* bits for which track is the left */
#define MASTER_TRACK 1
#define MASTER_BRANCH 2

/* bits for which MIDI events to duplicate and send to the branch */
#define DUPE_NOTE 1
#define DUPE_CONTROL 2
#define DUPE_CHANNEL 4

/* bits for which events to trash */
#define DISCARD_LR 1
#define DISCARD_DISTANCE 2
#define DISCARD_SYSEX 4
/* default tool settings */

#define DEFAULT_IN MIDI_Pan
#define DEFAULT_TRACK_OUT MIDI_Volume
#define DEFAULT_BRANCH_OUT MIDI_Volume
/* 0==off */
#define DEFAULT_ANGLE 0
#define DEFAULT_DISTANCE 0
#define DEFAULT_MASTER MASTER_TRACK
#define DEFAULT_TRASH (DISCARD_LR | DISCARD_DISTANCE | DISCARD_SYSEX)
#define DEFAULT_DUPE  DUPE_NOTE
#define DEFAULT_X 0.0
#define DEFAULT_Y 1.0
/* speaker angle from listener theta = 45 == MIDI 127 */
#define DEFAULT_THETA 0.785398163

#define TOOL_SIZE sizeof(struct SpaceTool) /* for ToolMaster */

#define GUP 100
#define GDOWN 200

/**** this poop is to use floating point in scm.lib w/ v6.5***/
void __regargs _CXBRK(void) { return; }        /* really no ^C checking */
void __stdargs _XCEXIT(long code) { return; }  /* declared in stdlib.h  */

extern struct Functions * far functions;


#if DBUG
static void __regargs dumptool(t)
struct SpaceTool *t;
{
	char s[20];

//	dprintf("in=%ld tout=%ld bout=%ld angle=%ld distance=%ld\n",t->in,t->tout,t->bout,t->angle,t->distance);
	dprintf("in=%ld tout=%ld bout=%ld distance=%ld\n",t->in,t->tout,t->bout,t->distance);
	sprintf(s,"%8.6f",t->x);
	dprintf("master=%ld dupe=%ld trash=%ld x=%s y=",t->master,t->dupe,t->trash,s);
	sprintf(s,"%8.6f",t->y);
	dprintf("%s ",s);
//	sprintf(s,"%8.6f",t->theta);
//	dprintf("theta=%s\n\n",s);
}
#endif

/***************************
	now, the guts of the Tool
****************************/

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(tool)

struct SpaceTool *tool;

{
	tool->tool.touched |= TOUCH_INIT;
	tool->in = DEFAULT_IN;
	tool->tout = DEFAULT_TRACK_OUT;
	tool->bout = DEFAULT_BRANCH_OUT;
	tool->angle = DEFAULT_ANGLE;
	tool->distance = DEFAULT_DISTANCE;
	tool->master = DEFAULT_MASTER;
	tool->trash = DEFAULT_TRASH;
	tool->dupe = DEFAULT_DUPE;
	tool->x = DEFAULT_X;
	tool->y = DEFAULT_Y;
	tool->theta = DEFAULT_THETA;
}

#define RIGHT_ANGLE 1.570796327
/*==========================================================================
* gain
*
*  INPUT: x & y coordinates for virtual sound location, angle between
*         speakers (in radians), speakers' distance from listener
* RETURN: gain factor 0.0 <= g >= 1.0
*
* EFFECT: ripped off from Moore, p358.  Calculate a gain factor based on 
*         the input goodies taking the inverse square law into account.
*=========================================================================*/

static double __regargs gain(x,y,angle,distance)

double x, y, angle, distance;

{
	double diff;
#if DBUG
 char s[16];
#endif

/*! WEENIE KLUDGE ALERT!!  because I'm too lame to figure out why this function
	returns 0 when x==0 && angle=135, and how to avoid /0 in the atan2() call,
	this avoids the issue entirely !!*/

/*?? shouldn't angle be tested against tool->theta, & diff adjusted accordingly?*/
	if (x == 0.0 && angle == 2.35619449) diff = 0.785398163;
	else {	
		/* normalize angle within range of atan() */
		angle = atan2(sin(angle),cos(angle));
		diff = fabs(angle - atan2(y,x));
	}
#if DBUG
 sprintf(s,"%8.6f",x);
 dprintf("gain:  x=%s ",s);
 sprintf(s,"%8.6f",y);
 dprintf("y=%s ",s);
 sprintf(s,"%13.9f",angle);
 dprintf("angle=%s ",s);
 sprintf(s,"%13.9f",diff);
 dprintf("diff=%s ",s);
 sprintf(s,"%8.6f",distance);
 dprintf("distance=%s\n",s);
#endif
	if (diff > RIGHT_ANGLE) return(0.0);
	else {
		diff = cos(diff) * distance / sqrt(x*x + y*y);
		return( (diff < 1.0) ? diff : 1.0 );
	}
//		return( min(1.0,cos(diff) * distance / sqrt(x*x + y*y)) );
}

/*==========================================================================
* hiloCont
*
*  INPUT: pointers to window and gadget, gadget value
* OUTPUT: none
* RETURN: gadget value
*
* EFFECT: display a pot value: MIDI continuous controller with an off=0 
*=========================================================================*/

static long __saveds __stdargs hiloCont(w,g,h)

struct Window *w;
struct Gadget *g;
long h;

{
	char text[4];

	SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset,
	         HILO_APEN,HILO_BPEN,HILO_DRAWMODE);
	if (h) sprintf(text,"%3ld",h-1);
	else sprintf(text,"%s","Off");
	Text(w->RPort,text,3);
	return(h);
}

#define tw tool->tool.window

/*==========================================================================
* drawpots
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: refreshes the pots
*=========================================================================*/

static void __regargs drawpots(tool)

struct SpaceTool *tool;

{
	functions->ModifyEmbossedProp(tw,LR_IN,tool->in,0,0,0,0,0);
	functions->ModifyEmbossedProp(tw,DISTANCE_IN,tool->distance,0,0,0,0,0);
//functions->ModifyEmbossedProp(tw,ANGLE,tool->angle,0,0,0,0,0);
	functions->ModifyEmbossedProp(tw,TRACK_OUT,tool->tout,0,0,0,0,0);
	functions->ModifyEmbossedProp(tw,BRANCH_OUT,tool->bout,0,0,0,0,0);
	functions->DrawEmbossedProp(tw,LR_IN);
	functions->DrawEmbossedProp(tw,DISTANCE_IN);
//functions->DrawEmbossedProp(tw,ANGLE);
	functions->DrawEmbossedProp(tw,TRACK_OUT);
	functions->DrawEmbossedProp(tw,BRANCH_OUT);
}


/*==========================================================================
* drawbool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: refreshes the boolean gadgets
*=========================================================================*/

static void __regargs drawbool(tool)

struct SpaceTool *tool;

{
	struct Gadget *g;

	functions->SelectEmbossed(tw,TRACK_LEFT,tool->master & MASTER_TRACK);
	functions->SelectEmbossed(tw,BRANCH_LEFT,tool->master & MASTER_BRANCH);
	functions->SelectEmbossed(tw,NOTES,tool->dupe & DUPE_NOTE);
	functions->SelectEmbossed(tw,CONTROL,tool->dupe & DUPE_CONTROL);
	functions->SelectEmbossed(tw,CHANNEL,tool->dupe & DUPE_CHANNEL);
	functions->SelectEmbossed(tw,TRASH_SYSEX,tool->trash & DISCARD_SYSEX);
	/* the two controller discard gadgets, with their own imagery */
	g = functions->GetGadget(tw,TRASH_LR);
	if (tool->trash & DISCARD_LR) g->Flags |= SELECTED;
	else g->Flags &= ~SELECTED;
	if (tool->trash & DISCARD_DISTANCE) g->NextGadget->Flags |= SELECTED;
	else g->NextGadget->Flags &= ~SELECTED;
	RefreshGList(g,tw,NULL,2);
}


/* scale MIDI 0..127 to 0..2 for Cartesian coordinates */
#define scale_y(b) ( ((b) * 0.015748031) + 1.0 )

/* scale MIDI 0..127 to 0..1 for Cartesian coordinates */
#define scale_x(b) ( ((b) / 63.0) - 1.0 )


/*==========================================================================
* dosysex
*
*  INPUT: pointer to event
* RETURN: pointer to event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static __inline struct Event *dosysex(tool,event)

struct SpaceTool *tool;
struct StringEvent *event;

{
	register unsigned char *sxdata;
	UBYTE refresh = 0;

	if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
#if DBUG
 dprintf("sysex=");
 for (refresh=0 ; (refresh<9) && (*(sxdata+refresh)!=MIDI_EOX) ; refresh++) 
 	dprintf("%3d ",*(sxdata+refresh));
 dprintf("\n");
 refresh=0;
#endif
		if (*sxdata != MIDI_EOX) {
			tool->x = scale_x(*sxdata);
			if (*(++sxdata) != MIDI_EOX) {
				tool->y = scale_y(*sxdata);
				if (*(++sxdata) != MIDI_EOX) {
					tool->master = *sxdata;
					refresh = 1; /* this is the first item to show in the ui */
					if (*(++sxdata) != MIDI_EOX) {
						tool->dupe = 0;
						if (*sxdata & DUPE_CHANNEL) tool->dupe = DUPE_CHANNEL;
						else {
							if (*sxdata & DUPE_CONTROL) tool->dupe = DUPE_CONTROL;
							if (*sxdata & DUPE_NOTE) tool->dupe |= DUPE_NOTE;
						}
						if (*(++sxdata) != MIDI_EOX) {
							tool->trash = *sxdata;
							if ( (*(++sxdata) != MIDI_EOX) && (*sxdata < 64) ) {
								tool->in = *sxdata;
								refresh = 3; /* this is the first pot in the ui */
								if ( (*(++sxdata) != MIDI_EOX) && (*sxdata < 65) ) {
									tool->distance = *sxdata; /* controller# + 1; 0 is off */
									if ( (*(++sxdata) != MIDI_EOX) && (*sxdata < 64) ) {
										tool->tout = *sxdata;
										if ( (*(++sxdata) != MIDI_EOX) && (*sxdata < 64) ) {
											tool->bout = *sxdata;
										}
									}
								}
							}
						}
					}
				}
			}
			if (refresh && RefreshOK(&tool->tool)) { /* refresh the gadgetry */
				if (refresh & 1) drawbool(tool);
				if (refresh & 2) drawpots(tool);
			}
		}
		return(NULL); /* can't free because of processevent() subsequent use */
	}
	return((struct Event *)event);
}

/*==========================================================================
* bround
*
*  INPUT: double
* RETURN: UBYTE
*
* EFFECT: returns the least erroneous unsigned byte 
*=========================================================================*/

static UBYTE __regargs bround(n)

double n;

{
	double nc, nf;

	nc = ceil(n);
	nf = floor(n);
	return( (UBYTE)((nc-n) <= (n-nf) ? nc : nf) );
}


/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to Tool
* RETURN: none
*
* EFFECT: builds sysex string from tool state and calls Snapshot(). adds returned
*         StringEvent to track's event list.
*=========================================================================*/

#define SYSEX_LENGTH 9

static void __inline dosnapshot(struct SpaceTool *tool)

{
	UBYTE sx[SYSEX_LENGTH];
	struct StringEvent *event;
	
	sx[0] = bround((tool->x + 1.0) * 63.0);
	sx[1] = bround((tool->y + 1.0) / 0.015748031);
	sx[2] = tool->master;
	sx[3] = tool->dupe;
	sx[4] = tool->trash;
	sx[5] = tool->in;
	sx[6] = tool->distance;
	sx[7] = tool->tout;
	sx[8] = tool->bout;
	if (event = BuildSnapshot(&tool->tool,
	                     (functions->running) ? functions->timenow : -1,
								sx, SYSEX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
	}
}

/*==========================================================================
* processevent
*
*  INPUT: pointer to Event
* RETURN: pointer to Event
*
* EFFECT: processes events one at a time
*=========================================================================*/

static struct Event * __saveds __stdargs processevent(event)

struct Event *event;

{
	register struct Event *newtrack;
	register struct Event *newbranch;
	struct Event *dupe = NULL;
	struct Event *trash = NULL;
	double mult;
	register struct SpaceTool *tool = (struct SpaceTool *)event->tool;
	UBYTE changed = 0;
#if DBUG
 char s[10];
#endif

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	event->tool = tool->tool.next;

//dprintf(" event @ %ld (%ld %ld %ld)\n",event,event->status,event->byte1,event->byte2);

	/* duplicate event if necessary and possible */
	if ((IsRealTime(event) && tool->dupe) || IsAllOff(event) || 
	     IsAllCtrlOff(event) ) {
		if (tool->tool.branch) { /* is there somewhere to send a duplicate? */
			/* is this something we need and are going to discard? */
			if ( ! (tool->trash && IsCtrlChg(event) &&
				  ((event->byte1 == tool->in) || (event->byte1 == tool->tout) ||
					 (event->byte1 == tool->bout) || /*(event->byte1 == tool->angle-1) ||*/
					 (event->byte1 == tool->distance-1)) ) ) {
				if (dupe = (struct Event *)(*functions->dupeevent)(event)) {
//dprintf("     dupe @ %ld\n",dupe);
					if ( (IsNote(event) && (tool->dupe & DUPE_NOTE)) ||
							(IsCtrlChg(event) && (tool->dupe & DUPE_CONTROL)) ||
							(IsChannelMsg(event) && (tool->dupe & DUPE_CHANNEL)) || 
							IsAllOff(event) || IsAllCtrlOff(event) ) {
						dupe->type |= EVENT_BRANCH;
						dupe->tool = tool->tool.branch;
					}
				}
			}
		}
	}

	if (event->type & EVENT_SYSX) {
		if (!(dosysex(tool,(struct StringEvent *)event))) {
			changed=1;
#if DBUG
 dprintf(" got sysex\n");
 dumptool(tool);
#endif
		}
#if DBUG
 else dprintf(" not our sysex\n");
#endif
	}

	if (IsRealTime(event) || changed) { /* look for changes, or sysex did already */
		if (!changed && IsCtrlChg(event)) {
			if (tool->distance && (event->byte1 == tool->distance-1)) {
				/* conjure another y coordinate */
				if ((mult=scale_y(event->byte2)) != tool->y) {
					tool->y = mult;
					changed = 1;
#if DBUG
 sprintf(s,"%9.6f",tool->y);
 dprintf("byte=%ld new y=%s\n",event->byte2,s);
#endif
				}
			}

			if (event->byte1==tool->in) {
				if ((mult=scale_x(event->byte2)) != tool->x) {
					tool->x = mult;
					changed=1;
#if DBUG
 sprintf(s,"%9.6f",tool->x);
 dprintf("byte=%ld new x=%s\n",event->byte2,s);
#endif
				}
			}

/***
		/* divine new speaker-spread angle */
			if (tool->angle && (event->byte1 == tool->angle-1)) {
				/* scale & convert new speaker-spread angle to radians */
				tool->theta = event->byte2 * 0.708661417; /* 127 <=> 90 */
#if DBUG
 sprintf(s,"%9.6f",tool->theta);
 dprintf("byte=%ld new angle=%s ",event->byte2,s);
#endif
				tool->theta = (PI * tool->theta) / 180.0;
				changed = 1;
#if DBUG
 sprintf(s,"%9.6f",tool->theta);
 dprintf("new theta=%s\n",s);
#endif
			}
***/
		} /* if IsCtrlChg() */

		if (changed) { /* new poop; recalculate gain for each channel */
/*!!*/			if (tool->trash) trash=event;
			newtrack = (struct Event *)functions->dupeevent(event);
			if (newtrack) {
				newtrack->type = EVENT_VOICE; /* in case the change was a sysex */
				newtrack->data = 0;
				newtrack->status = MIDI_CCHANGE;
				newtrack->byte1 = tool->tout;
				/* calculate gain for new location */
				mult = gain( tool->x, tool->y,
				             ((tool->master & MASTER_TRACK) ? 2.35619449 : 0.785398163),
										 1.414213562 ); /* distance to speaker fixed; angle can't change */
				newtrack->byte2 = bround(127.0*mult);
#if DBUG
// dumptool(tool);
 sprintf(s,"%9.6f",mult);
 dprintf("track gain=%s  newtrack->byte2=%ld\n",s,newtrack->byte2);
#endif
				if (tool->trash) {
					newtrack->next = event->next;
				}
				else {
					newtrack->next = event;
				}
				event = newtrack;
//dprintf(" newtrack @ %ld\n",newtrack);
				/* now do the branch, if there is one connected */
				if (tool->tool.branch) {
					newbranch = (struct Event *)(*functions->dupeevent)(newtrack);
					if (newbranch) {
						newbranch->type |= EVENT_BRANCH;
						newbranch->byte1 = tool->bout;
						mult = gain( tool->x, tool->y,
			      	           ((tool->master & MASTER_BRANCH) ? 2.35619449 : 0.785398163),
								         1.414213562 );
						newbranch->byte2 = bround(127.0*mult);
#if DBUG
 sprintf(s,"%9.6f",mult);
 dprintf("branch gain=%s newbranch->byte2=%ld\n",s,newbranch->byte2);
#endif
						newbranch->tool = tool->tool.branch;
						newbranch->next = event;
						event = newbranch;
//dprintf("newbranch @ %ld\n",newbranch);
					}
				} /* if (branch) */
			} /* if (newtrack) */
		} /* if (changed) */
#if DBUG
 dprintf("\n");
#endif
		/* if we're supposed to, and it was ours, and it's not linked in */
		if (tool->trash && changed && trash) {
//dprintf("trash @ %ld\n",trash);
			if (trash->type & EVENT_SYSX) {
				functions->freestring((struct String *)trash->data);
#if DBUG
 dprintf("free sysex\n");
#endif
			}
			functions->freeevent(trash);
		}
		if (dupe) { /* got another to stick in here? */
			dupe->next = event; /* put it at the head of the list */
			event = dupe;
		}
	} /* if IsRealTime() */
//dprintf("RETURNED LIST\n");
//for (trash=event ; trash ; trash=trash->next) dprintf("@ %X ",trash);
//dprintf("\n");
	return((struct Event *)event);
}

/* refresh bits */
#define REF_BOOL 1
#define REF_POTS 2

/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __saveds __stdargs edittool(tool)

struct SpaceTool *tool;

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	UBYTE refresh;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

	nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,20);
	if (!nw) return;
	AddSnapshotGadget(tw);
	SetMenuStrip(tw,&Menu1);
	functions->FatEmbossedPropOn(tw,LR_IN,LR_IN+GDOWN,LR_IN+GUP,hilo,64,1);
	functions->FatEmbossedPropOn(tw,TRACK_OUT,TRACK_OUT+GDOWN,TRACK_OUT+GUP,hilo,64,1);
	functions->FatEmbossedPropOn(tw,BRANCH_OUT,BRANCH_OUT+GDOWN,BRANCH_OUT+GUP,hilo,64,1);
//	functions->FatEmbossedPropOn(tw,ANGLE,ANGLEL,ANGLER,hiloCont,65,1);
	functions->FatEmbossedPropOn(tw,DISTANCE_IN,DISTANCE_IN+GDOWN,DISTANCE_IN+GUP,hiloCont,65,1);
	EmbossOnGList(tw,&Gadget1,GCOUNT,1);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	refresh = REF_BOOL | REF_POTS;
	SetRefreshOK(&tool->tool);

	/*** ui handling ***/
	for ( ; ; ) {
		if (refresh) {
#if DBUG
 dumptool(tool);
#endif
			if (refresh & REF_POTS) drawpots(tool);
			if (refresh & REF_BOOL) drawbool(tool);
			refresh=0;
		}
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			ResetRefreshOK(&tool->tool);
			break;
		}
		else if (class == GADGETUP) {
			if (gadget->GadgetID == SNAPSHOT_GADGET_ID) {
				dosnapshot(tool);
			}
			if (gadget->GadgetID == TRACK_LEFT) {
				tool->master = MASTER_TRACK;
			}
			else if (gadget->GadgetID == BRANCH_LEFT) {
				tool->master = MASTER_BRANCH;
			}
			else if (gadget->GadgetID == NOTES) { /* excludes all and channel*/
				if (tool->dupe & DUPE_CHANNEL) 
					tool->dupe = DUPE_NOTE;
				else tool->dupe ^= DUPE_NOTE;
			}
			else if (gadget->GadgetID == CONTROL) { /* excludes channel and all */
				tool->dupe = (tool->dupe & (DUPE_NOTE|DUPE_CONTROL))^DUPE_CONTROL;
			}
			else if (gadget->GadgetID == CHANNEL) { /* excludes everything else */
				if (tool->dupe & DUPE_CHANNEL) tool->dupe = 0;
				else tool->dupe = DUPE_CHANNEL;
			}
			else if (gadget->GadgetID == TRASH_SYSEX) {
				tool->trash ^= DISCARD_SYSEX;
			}
			else if (gadget->GadgetID == TRASH_LR) {
				tool->trash ^= DISCARD_LR;
			}
			else if (gadget->GadgetID == TRASH_DISTANCE) {
				tool->trash ^= DISCARD_DISTANCE;
			}
			refresh=REF_BOOL;
		}
		else if (class == GADGETDOWN) {
			switch (gadget->GadgetID) {
				case LR_IN : 
					tool->in = functions->DragEmbossedProp(tw,LR_IN);
					refresh=REF_POTS;
					break;
				case LR_IN+GDOWN :
					tool->in = functions->ShiftEmbossedProp(tw,LR_IN,-1,0);
					refresh=REF_POTS;
					break;
				case LR_IN+GUP :
					tool->in = functions->ShiftEmbossedProp(tw,LR_IN,1,0);
					refresh=REF_POTS;
					break;
				case TRACK_OUT : 
					tool->tout = functions->DragEmbossedProp(tw,TRACK_OUT);
					refresh=REF_POTS;
					break;
				case TRACK_OUT+GDOWN :
					tool->tout = functions->ShiftEmbossedProp(tw,TRACK_OUT,-1,0);
					refresh=REF_POTS;
					break;
				case TRACK_OUT+GUP :
					tool->tout = functions->ShiftEmbossedProp(tw,TRACK_OUT,1,0);
					refresh=REF_POTS;
					break;
				case BRANCH_OUT : 
					tool->bout = functions->DragEmbossedProp(tw,BRANCH_OUT);
					refresh=REF_POTS;
					break;
				case BRANCH_OUT+GDOWN :
					tool->bout = functions->ShiftEmbossedProp(tw,BRANCH_OUT,-1,0);
					refresh=REF_POTS;
					break;
				case BRANCH_OUT+GUP :
					tool->bout = functions->ShiftEmbossedProp(tw,BRANCH_OUT,1,0);
					refresh=REF_POTS;
					break;
/*
				case ANGLE : 
					tool->angle = functions->DragEmbossedProp(tw,ANGLE);
					refresh=REF_POTS;
					break;
				case ANGLEL :
					tool->angle = functions->ShiftEmbossedProp(tw,ANGLE,-1,0);
					refresh=REF_POTS;
					break;
				case ANGLER :
					tool->angle = functions->ShiftEmbossedProp(tw,ANGLE,1,0);
					refresh=REF_POTS;
					break;
*/
				case DISTANCE_IN : 
					tool->distance = functions->DragEmbossedProp(tw,DISTANCE_IN);
					refresh=REF_POTS;
					break;
				case DISTANCE_IN+GDOWN :
					tool->distance = functions->ShiftEmbossedProp(tw,DISTANCE_IN,-1,0);
					refresh=REF_POTS;
					break;
				case DISTANCE_IN+GUP :
					tool->distance = functions->ShiftEmbossedProp(tw,DISTANCE_IN,1,0);
					refresh=REF_POTS;
					break;
			}
		}
	}
	functions->FatEmbossedPropOff(tw,TRACK_OUT,TRACK_OUT+GDOWN,TRACK_OUT+GUP);
	functions->FatEmbossedPropOff(tw,BRANCH_OUT,BRANCH_OUT+GDOWN,BRANCH_OUT+GUP);
//	functions->FatEmbossedPropOff(tw,ANGLE,ANGLEL,ANGLER);
	functions->FatEmbossedPropOff(tw,DISTANCE_IN,DISTANCE_IN+GDOWN,DISTANCE_IN+GUP);
	EmbossOffGList(tw,&Gadget1,GCOUNT);
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* shutdown
*
*  INPUT: pointer to tool, current time
* RETURN: none
*
* EFFECT: resets the tool's output controllers to known states
*=========================================================================*/

static __inline void shutdown(tool,time)

struct SpaceTool *tool;
long time;

{
	struct Event *new1;
	struct Event *new2;

	new1 = (struct Event *)functions->allocevent();
	if (new1) {
		new1->type = EVENT_VOICE;
		new1->time = time;
		new1->status = MIDI_CCHANGE;
		new1->byte1 = tool->tout;
		new1->byte2 = 127;
		new1->tool = tool->tool.next;
		functions->qevent(new1);
		if (tool->tool.branch) {
			new2 = (struct Event *)functions->dupeevent(new1);
			if (new2) {
				new2->type |= EVENT_BRANCH;
				new2->byte1 = tool->bout;
				new2->tool = tool->tool.branch;
				functions->qevent(new2);
			}
		}
	}
	tool->x = DEFAULT_X;
	tool->y = DEFAULT_Y;
}


struct ToolMaster master;

/*==========================================================================
* transportcode
*
*  INPUT: command, current time, end time
* RETURN: none
*
* EFFECT: if the command is a stop, look for all instances of this tool and 
*         shut 'em down
*=========================================================================*/

static void __saveds __stdargs transportcode(command,time,end)

long command, time, end;

{
	struct SpaceTool *tool;
	struct Track *track;

	if (command != TC_STOP) return;

	if (functions->multiin) track = (struct Track *)functions->tracklist;
	else track = master.intrack;
	for ( ; track ; track = track->next ) {
		tool = (struct SpaceTool *)track->toollist;
		for ( ; tool ; tool=(struct SpaceTool *)tool->tool.next) {
			if (tool->tool.toolid == THIS_TOOL_ID) shutdown(tool,time);
		}
	}
}

/*********************************/
/*	Remove the transport handler. */

static void __saveds __stdargs removetool(void)

{
	functions->removetransport(transportcode);
}

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to tool master struct
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __saveds __stdargs inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = THIS_TOOL_ID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.image = &SpaceIconDown;
	master.upimage = &SpaceIconUp;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL|TOOL_BRANCHOUT;
	master.removetool = removetool;
	functions->installtransport(transportcode);
	return(&master);
}
